using System;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using Microsoft.SPOT;
using Microsoft.SPOT.Hardware;
using SecretLabs.NETMF.Hardware;
using SecretLabs.NETMF.Hardware.Netduino;

namespace I2C
{
    /// <summary>
    /// This class is a base class for classes implementing interfaces to I2C  capable devices.
    /// This class is not meant to be instatiated directly.
    /// </summary>
    public class I2CBaseSlave
    {
        #region Attributes
        protected I2CDevice.Configuration DeviceConfig;
        protected int TransactionTimeout { get; set; }        // Stored in milliseconds
        protected int ClockRateKHz { get; set; }
        protected ushort Address { get; set; }
#if (MICROSOFTWAY)
        protected I2CDevice Device;
#else
        protected static TechWithDave.SoftwareI2CBus I2CBus = null;
        protected TechWithDave.SoftwareI2CBus.I2CDevice Device;
#endif
        #endregion Attributes


        public static void SetPins(Cpu.Pin scl = Cpu.Pin.GPIO_NONE, Cpu.Pin sda = Cpu.Pin.GPIO_NONE)
        {
            if (I2CBus == null)
            {
                if (scl == Cpu.Pin.GPIO_NONE)
                {
                    scl = Pins.GPIO_PIN_SCL;
                }
                if (sda == Cpu.Pin.GPIO_NONE)
                {
                    sda = Pins.GPIO_PIN_SDA;
                }
                I2CBus = new TechWithDave.SoftwareI2CBus(scl, sda);
            }
        }
        
        
        /// <summary>
        /// class constructor
        /// </summary>
        /// <param name="address">Address of device</param>
        /// <param name="clockRate">The clock rate in kHz</param>
        /// <param name="timeout">The timeout for each operation in mS</param>
        protected I2CBaseSlave(ushort address, int clockRate = 100, int timeout = 3000)
        {
            Address = address;
            ClockRateKHz = clockRate;
            TransactionTimeout = timeout;
            DeviceConfig = new I2CDevice.Configuration(Address, ClockRateKHz);
#if (MICROSOFTWAY)
            Device = new I2CDevice(DeviceConfig);
#else
            if (I2CBus == null)
                SetPins();
            Device = new TechWithDave.SoftwareI2CBus.I2CDevice(I2CBus, (byte)DeviceConfig.Address, (int)DeviceConfig.ClockRateKhz);
#endif
        }

        /// <summary>
        /// Write the data in the buffer to the bus
        /// </summary>
        /// <param name="data">The buffer</param>
        protected int Write (byte[] data)
        {
            I2CDevice.I2CWriteTransaction [] transaction = new I2CDevice.I2CWriteTransaction[] {I2CDevice.CreateWriteTransaction(data)};
            return Device.Execute(transaction, TransactionTimeout);
        }

        /// <summary>
        /// Write the register address to the bus
        /// </summary>
        /// <param name="register">The register to write to (device dependent)</param>
        protected void WriteRegister(byte register)
        {
            byte[] data = new byte[1] { register };
            Write(data);
        }

        /// <summary>
        /// Write data to a register
        /// </summary>
        /// <param name="register">The register to write to (device dependent)</param>
        /// <param name="data">The data to write</param>
        /// <param name="timeoutms">The timeout in milliseconds</param>
        protected void WriteRegister(byte register, byte value)
        {
            byte[] data = new byte[2] {register, value };
            Write(data);
        }

        /// <summary>
        /// Read the specified number of bytes into the buffer
        /// </summary>
        /// <param name="data">The buffer to read into</param>
        /// <returns>The buffer</returns>
        protected byte[] Read (byte[] data)
        {
            I2CDevice.I2CReadTransaction[] transaction = new I2CDevice.I2CReadTransaction[] {I2CDevice.CreateReadTransaction(data)};
            Device.Execute(transaction, TransactionTimeout);
            return data;
        }

        /// <summary>
        /// Reads the perscribed amount of data from the given register
        /// </summary>
        /// <param name="register">The register to read from</param>
        /// <param name="data">The buffer to read into, buffer size specifies amount of data to read</param>
        /// <returns>The buffer</returns>
        protected byte[] ReadRegister(byte register, byte[] data)
        {
            WriteRegister(register);
            data = Read(data);
            return data;
        }
    }
}

