using System.Threading;
using Microsoft.SPOT;
using Microsoft.SPOT.Hardware;
using SecretLabs.NETMF.Hardware.Netduino;

namespace I2C
{
    class LCDDisplay : I2CBaseSlave
    {
        public static void TestProgram()
        {
            OutputPort VCC = new OutputPort(Pins.GPIO_PIN_D13, true);
            LCDDisplay lcd = new LCDDisplay(0x3f, 0, 2, 1, 0, 4, 5, 6, 7, 3);
            while (true)
            {
                lcd.BackLightOn();
                lcd.SetPosition(1, 0);
                lcd.WriteString("Hello World!");
                lcd.SetPosition(2, 0);
                lcd.WriteString("Is this working...");
                Thread.Sleep(3000);
                lcd.Clear();
            }
        }


        public enum Registers
        {
              CMD_Clear_Display = 0x01
            , CMD_Return_Home = 0x02
            , CMD_Entry_Mode = 0x04
            , CMD_Display_Control = 0x08
            , CMD_Cursor_Display_Shift = 0x10
            , CMD_Function_Set = 0x20
            , CMD_DDRAM_Set = 0x80
        };
        public enum RegisterOptions
        {
              OPT_Increment = 0x02 					// CMD_Entry_Mode
            , OPT_Display_Shift = 0x01 				// CMD_Entry_Mode
            , OPT_Enable_Display = 0x04 			// CMD_Display_Control
            , OPT_Enable_Cursor = 0x02 				// CMD_Display_Control
            , OPT_Enable_Blink = 0x01 				// CMD_Display_Control
            , OPT_Shift_Display = 0x08 				// CMD_Cursor_Display_Shift
            , OPT_Shift_Right = 0x04 				// CMD_Cursor_Display_Shift 0 = Left
            , OPT_2_Lines = 0x08 					// CMD_Function_Set 0 = 1 line
            , OPT_5x10_Dots = 0x04                  // CMD_Function_Set 0 = 5x7 dots
        };

        private enum EPins { VSS, VDD, V0, RS, RW, E, D0, D1, D2, D3, D4, D5, D6, D7, A, K, Backlight, Size };
        private enum EOnOff { On, Off };
        private int[] _Pins = new int[(int)EPins.Size];
        private EOnOff _BacklightState = EOnOff.Off;

        public LCDDisplay(int addr, int port, int en, int rw, int rs, int d4, int d5, int d6, int d7, int backlight = -1)
            : base((ushort)addr, 400)
        {
            _Pins[(int)EPins.E] = en;
            _Pins[(int)EPins.RS] = rs;
            _Pins[(int)EPins.RW] = rw;
            _Pins[(int)EPins.D4] = d4;
            _Pins[(int)EPins.D5] = d5;
            _Pins[(int)EPins.D6] = d6;
            _Pins[(int)EPins.D7] = d7;
            _Pins[(int)EPins.Backlight] = backlight;
            _BacklightState = EOnOff.Off;

            // Activate LCD
            byte initialize_i2c_data = 0x00;
            initialize_i2c_data = _pinInterpret(EPins.D4, initialize_i2c_data, true);
            initialize_i2c_data = _pinInterpret(EPins.D5, initialize_i2c_data, true);
            _enable(initialize_i2c_data);
            Thread.Sleep(200);
            _enable(initialize_i2c_data);
            Thread.Sleep(100);
            _enable(initialize_i2c_data);
            Thread.Sleep(100);

            // Initialize 4-bit mode
            initialize_i2c_data = _pinInterpret(EPins.D4, initialize_i2c_data, false);
            _enable(initialize_i2c_data);
            Thread.Sleep(10);

            command((byte)Registers.CMD_Function_Set | (byte)RegisterOptions.OPT_2_Lines);
            command((byte)Registers.CMD_Display_Control | (byte)RegisterOptions.OPT_Enable_Display | (byte)RegisterOptions.OPT_Enable_Cursor);
            command(Registers.CMD_Clear_Display);
            command((byte)Registers.CMD_Entry_Mode | (byte)RegisterOptions.OPT_Increment | (byte)RegisterOptions.OPT_Display_Shift);
        }


        public void Clear()
        {
            command(Registers.CMD_Clear_Display);
            Thread.Sleep(100);
        }

        public void home()
        {
            command(Registers.CMD_Return_Home);
            Thread.Sleep(100);
        }

        public void SetPosition(int line, int pos)
        {
            Debug.Print("Set Position : " + line.ToString() + ", " + pos.ToString());
            int[] constModifier = { 0x0, 0x40, 0x14, 0x54 };
            if (line <= 4 && line > 0)
            {
                int address = constModifier[line-1] + pos;
                command((int)Registers.CMD_DDRAM_Set + address);
            }
        }

        public void WriteChar(char data)
        {
            _write((byte)data, false);
        }

        public void WriteString(string data)
        {
            Debug.Print("Writing : " + data);
            foreach (char c in data)
            {
                WriteChar(c);
            }
        }

        private void BackLight(bool newState)
        {
            if (_Pins[(int)EPins.Backlight] >= 0)
            {
                Write(new byte[] { _pinInterpret(EPins.Backlight, 0x00, newState) });
                _BacklightState = (newState ? EOnOff.On : EOnOff.Off);
            }
        }

        public void BackLightOn() { BackLight(true); }
        public void BackLightOff() { BackLight(false); }


        /// <summary>
        /// Write a byte of data/command to the display
        /// </summary>
        /// <param name="data">The data to write</param>
        /// <param name="command">True if this is a command, false if its data</param>
        private void _write(byte data, bool command = true)
        {
            byte hi_nibble = (byte)(data >> 4);
            byte lo_nibble = (byte)(data & 0x0F);
            WriteNibble(hi_nibble, command);
            WriteNibble(lo_nibble, command);
            Thread.Sleep(10);
        }
        /// <summary>
        /// Write half of the data that needs to be written.
        /// </summary>
        /// <param name="nibble">The data to be written</param>
        /// <param name="command">True if this is a command false if it is data</param>
        private void WriteNibble(byte nibble, bool command)
        {
            byte i2c_data = 0x00;
            i2c_data = _pinInterpret(EPins.D4, i2c_data, ((nibble & 0x01) != 0));
            i2c_data = _pinInterpret(EPins.D5, i2c_data, (((nibble >> 1) & 0x01) != 0));
            i2c_data = _pinInterpret(EPins.D6, i2c_data, (((nibble >> 2) & 0x01) != 0));
            i2c_data = _pinInterpret(EPins.D7, i2c_data, (((nibble >> 3) & 0x01) != 0));

            if (!command)		    // Set the register selector to 1 if this is data
            {
                i2c_data = _pinInterpret(EPins.RS, i2c_data, true);
            }
            _enable(i2c_data);		    // Toggle Enable 
        }

        private byte _pinInterpret(EPins pin, byte data, bool value = false)
        {
            if (value)
            {
                // Construct mask using pin
                byte mask = (byte)(0x01 << _Pins[(int)pin]);
                data = (byte)(data | mask);
            }
            else
            {
                // Construct mask using pin
                byte mask = (byte)((0x01 << _Pins[(int)pin]) ^ 0xFF);
                data = (byte)(data & mask);
            }
            return data;
        }

        private void _enable(byte data)
        {
            // Determine if black light is on and insure it does not turn off or on
            if (_BacklightState == EOnOff.On)
            {
                data = _pinInterpret(EPins.Backlight, data, true);
            }
            else
            {
                data = _pinInterpret(EPins.Backlight, data, false);
            }
            Write(new byte[] { data });
            Write(new byte[] { _pinInterpret(EPins.E, data, true) });
            Write(new byte[] { data });
        }
        // For legacy
        private void command(byte data) { _write(data, true); }
        private void command(int data) { _write((byte)data, true); }
        private void command(Registers command) { _write((byte)command, true); }
    }
}
