// SoftwareI2CBus.cs -- Wrapper for SoftwareI2C to more closely emulate 
// Microsoft.SPOT.Hardware.I2CDevice and support multiple I2C buses.  
//
// SoftwareI2CBus.cs License: Public Domain
// No warranty for use.
//
// Note: Requires separate SoftwareI2C class posted at http://www.tinyclr.com/codeshare/entry/236

using System;
using GHIElectronics.NETMF.Hardware.Emulation; // for SoftwareI2C class
using Microsoft.SPOT.Hardware;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using Microsoft.SPOT;
using SecretLabs.NETMF.Hardware;
using SecretLabs.NETMF.Hardware.Netduino;

namespace TechWithDave
{
    public class SoftwareI2CBus
    {
        public SoftwareI2CBus(Cpu.Pin scl, Cpu.Pin sda)
        {
            Scl = scl;
            Sda = sda;
        }

        public Cpu.Pin Scl 
        { 
            get; 
            private set;
        }

        public Cpu.Pin Sda
        {
            get;
            private set;
        }

        /// <summary>
        /// Create a new device on the bus
        /// </summary>
        /// <param name="config"></param>
        /// <returns></returns>
        public I2CDevice CreateI2CDevice(Microsoft.SPOT.Hardware.I2CDevice.Configuration config)
        {
            return new I2CDevice(this, (byte)config.Address, config.ClockRateKhz); // note: currently only supports 7-bit addresses
        }

        /// <summary>
        /// This class represents a device on the above bus
        /// </summary>
        public class I2CDevice
        {
            private SoftwareI2C i2c;
            byte addr;

            internal I2CDevice(SoftwareI2CBus i2cbus, byte addr, int kHz)
            {
                i2c = new SoftwareI2C(i2cbus.Scl, i2cbus.Sda, kHz);
                this.addr = addr;
            }

            /// <summary>
            /// Execute a stream of operations
            /// </summary>
            /// <param name="transactions"></param>
            /// <param name="timeoutms"></param>
            /// <returns></returns>
            public int Execute(Microsoft.SPOT.Hardware.I2CDevice.I2CTransaction[] transactions, int timeoutms)
            {
                int total = 0;

                foreach (Microsoft.SPOT.Hardware.I2CDevice.I2CTransaction transaction in transactions)
                {
                    int count;

                    if (transaction is Microsoft.SPOT.Hardware.I2CDevice.I2CReadTransaction)
                        count = Execute((Microsoft.SPOT.Hardware.I2CDevice.I2CReadTransaction)transaction);
                    else if (transaction is Microsoft.SPOT.Hardware.I2CDevice.I2CWriteTransaction)
                        count = Execute((Microsoft.SPOT.Hardware.I2CDevice.I2CWriteTransaction)transaction);
                    else
                        throw new NotImplementedException(transaction.GetType().ToString());

                    if (count == 0)
                        break;

                    total += count;
                }

                return total;
            }

            /// <summary>
            /// Execute a read operation
            /// </summary>
            /// <param name="transaction"></param>
            /// <returns></returns>
            public int Execute(Microsoft.SPOT.Hardware.I2CDevice.I2CReadTransaction transaction)
            {
                int count = 0;

                if (!i2c.Transmit(true, false, (byte)((addr << 1) | 0x01)))
                {
                    while (count < transaction.Buffer.Length)
                    {
                        byte data = i2c.Receive(true, count == transaction.Buffer.Length-1);
                        transaction.Buffer[count++] = data;
                    }
                }

                return count;
            }

            /// <summary>
            /// Execute a single write operation
            /// </summary>
            /// <param name="transaction"></param>
            /// <returns></returns>
            public int Execute(Microsoft.SPOT.Hardware.I2CDevice.I2CWriteTransaction transaction)
            {
                int count = 0;

                if (!i2c.Transmit(true, false, (byte)(addr << 1)))
                {
                    while (count < transaction.Buffer.Length)
                    {
                        if (i2c.Transmit(false, count == transaction.Buffer.Length - 1, transaction.Buffer[count]))
                            break;
                        else
                            ++count;
                    }
                }

                return count;
            }

            /// <summary>
            /// Read data from the bus
            /// </summary>
            /// <param name="opCode"></param>
            /// <param name="address"></param>
            /// <param name="data"></param>
            /// <returns></returns>
            public byte[] Read(byte opCode, byte address, byte[] data)
            {
                const byte constReadFlag = 0x01;
                int count = 0;
                // The sequence is S OpCode R Addr DOut P
                if (!i2c.Transmit(true, false, opCode))                 
                {
                    if (!i2c.Transmit(false, false, (byte)((addr << 1) | constReadFlag)))
                    {
                        while (count < data.Length)
                        {
                            byte rxedByte = i2c.Receive(true, count == data.Length - 1);
                            data[count++] = rxedByte;
                        }
                    }
                }
                return data;
            }

            /// <summary>
            /// Write data from the bus
            /// </summary>
            /// <param name="opCode"></param>
            /// <param name="address"></param>
            /// <param name="data"></param>
            public void Write(byte opCode, byte address, byte[] data)
            {
                const byte constWriteFlag = 0x0;
                int count = 0;
                // The sequence is S OpCode R Addr DOut P
                if (!i2c.Transmit(true, false, opCode))
                {
                    if (!i2c.Transmit(false, false, (byte)((addr << 1) | constWriteFlag)))
                    {
                        while (count < data.Length)
                        {
                            if (i2c.Transmit(false, count == data.Length - 1, data[count]))
                                break;
                            else
                                ++count;
                        }
                    }
                }
            }
        }
    }
}
